#!/usr/bin/env python
#
#
#   Copyright (c) 2007 Jesse Guardiani <jesse@guardiani.us>
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License as
#   published by the Free Software Foundation; either version 2 of the
#   License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
#   02111-1307, USA.
#

import pygame, time
from pygame.locals import *
import globals
from db      import db      as db
from theme   import theme   as theme
from manager import manager as manager


class BaseWidget(pygame.sprite.Sprite):
  def __init__(self):
    pygame.sprite.Sprite.__init__(self) # call Sprite initializer
    self.image = None
    self.rect  = None

  def handle_event(self,event):
    ''' override me '''

  def is_idle(self):
    ''' override me '''
    return 1

  def update(self):
    ''' override me '''

  def draw(self,surface):
    surface.blit(self.image,self.rect)

  def close(self):
    ''' override me '''


class Button(BaseWidget):
#
# we could also add a is_escape=False for albumbuttons etc below, which would then binds to the esc key
# but is it the right approach? -disq
#
  def __init__(self,image,onclick_cb,is_default=False):
    BaseWidget.__init__(self)
    self.image = image
    self.rect  = self.image.get_rect()
    self.onclick_cb = onclick_cb
    self.mousedown = False
    self.mouseup_timeout_start = None
    self.is_default = is_default

  def handle_event(self,event):
    if event.type == MOUSEBUTTONDOWN:
      # only register a button click if there has been some time since last button click. This prevents false clicks.
      if not self.mouseup_timeout_start or time.time()-self.mouseup_timeout_start > 0:
        self.mousedown = True
        self.mouseup_timeout_start = None
    if self.mousedown and event.type == MOUSEBUTTONUP:
      if self.rect.collidepoint(pygame.mouse.get_pos()):
        self.mousedown = False
        self.mouseup_timeout_start = time.time()
        self.onclick_cb()
    if self.is_default and event.type == KEYDOWN and event.key == K_RETURN:
        self.onclick_cb()


class PlayBackButton(Button):
  ''' like a Button, but only visible when playlist is not empty '''

  def handle_event(self,event):
    if not manager.playlist.list: return
    Button.handle_event(self,event)

  def draw(self,surface):
    if not manager.playlist.list: return
    Button.draw(self,surface)


class ScrollWidget(BaseWidget):
  ''' This widget wraps sprites with equal heights and 
  presents the user with an interface for scrolling and
  selecting these sprites'''
  def __init__(self,rect,selected_cb):
    BaseWidget.__init__(self) # call Group initializer
    self.rect       = pygame.Rect(rect) # this is where we draw our widget within the surface
    self.y_spacer   = 5 # height in pixels of whitespace between the top and bottom of sprites
    self.offset     = 0
    self.distance   = 0
    self.mousedown  = 0
    self.speed      = 0
    self.num_events = 0
    self.friction   = 2
    self.friction_c = 0
    self.selection_indicator = Button(theme.get_image('sel_indicator'),self._make_selection_indicator_callback())
    self.selection_indicator.rect.topleft = (-100,-100) # off screen at first
    self.viewport_index = 0
    self.is_scrollable  = None
    self.sprites        = []
    self.sorted_sprites = []
    self.selected_index = None
    self.need_sort      = False
    self.selected_cb    = selected_cb
    self.last_state     = "idle" # see self._change_state() for possibilities
    self.cur_state      = "idle"
    self.tap_event_threshold = 12
    self.tap_distance_threshold = 12


  def _make_selection_indicator_callback(self):
    def _callback():
      ''' do nothing '''
      return 
    return _callback

  def _change_state(self,new_state):
    l = ['idle','tap','drag']
    if new_state not in l:
      print "WARNING: unknown state %s" % (new_state,)
      return
    self.last_state = self.cur_state
    self.cur_state  = new_state

  def _sort_sprites(self):
    if (self.need_sort): # cache sort order between list updates
      self.is_scrollable = None
      print "sorting"
      self.need_sort = False
      self.sorted_sprites = self.sprites
      self.sorted_sprites.sort()

  def add(self,*sprite):
    self.need_sort = True
    self.sprites.append(*sprite)

  def remove(self,*sprite):
    self.need_sort = True
    self.sprites.remove(*sprite)

  def has_selected(self):
    ''' The "selected" sprite is the one that we think the user has chosen '''
    if self.selected_index == None: return False
    return True

  def next(self):
    ''' Move selected_sprite to next sprite in list '''
    if self.is_idle():
      if self.has_selected():
        self.set_selected_index(self.selected_index + 1,True)
      else:
        self.new_selected_index()

  def prev(self):
    ''' Move selected_sprite to previous sprite in list '''
    if self.is_idle():
      if self.has_selected():
        self.set_selected_index(self.selected_index - 1,True)
      else:
        self.new_selected_index()

  def get_selected_sprite(self):
    if not self.has_selected(): return None
    return self.sorted_sprites[self.selected_index]

  def _calc_new_index(self,index,list):
    length = len(list)
    while index != 0: # FIXME: looping is silly for very large abs() values of index. maybe use divmod instead?
      if index < 0: index = index + length
      elif index > length - 1: index = index - length
      else: break # within the normal range, so return it
    return index

  def new_selected_index(self):
    self.selected_index = self.viewport_index
    self.offset = 0
    self.update()

  def set_selected_index(self,index,snap=False):
    if self.has_selected(): self.get_selected_sprite().set_focus(False)
    self.selected_index = self._calc_new_index(index,self.sorted_sprites)
    print "selected=%s" % (self.selected_index,)
    self.get_selected_sprite().set_focus(True)
    if snap and self.is_scrollable:
      if not self.is_index_mostly_visible(self.selected_index):
        print "here"
        self.viewport_index = self.selected_index
        self.offset = 0
    self.update()

  def set_selected_sprite(self,new_selected_sprite):
    print "select it"
    c = 0
    for sprite in self.sorted_sprites: # make sure this new selected sprite exists in our list
      if new_selected_sprite == sprite:
        self.set_selected_index(c,True)
        return True
      c += 1
    print "invalid set_selected!" # this should never happen
    return False

  def get_sprites(self):
    self._sort_sprites()
    return self.sorted_sprites

  def _calc_viewport_index(self):
    while self.offset != 0:
      height = self.sorted_sprites[self.viewport_index].rect.height + self.y_spacer
      if self.offset > 0:
        if self.offset > height:
          self.offset = self.offset - height
          self.viewport_index = self._calc_new_index(self.viewport_index - 1,self.sorted_sprites)
        elif self.sorted_sprites[self.viewport_index].rect.top > self.rect.top:
          self.offset = 0 - height + self.offset
          self.viewport_index = self._calc_new_index(self.viewport_index - 1,self.sorted_sprites)
        else: return
      else:
        if abs(self.offset) > height:
          self.offset = self.offset + height
          self.viewport_index = self._calc_new_index(self.viewport_index + 1,self.sorted_sprites)
        else: return

  def _apply_friction(self):
    self.friction_c = self.friction_c + 1
    if self.friction_c >= self.friction:
      self.friction_c = 0
      self.speed = self.speed - cmp(self.speed,0)

  def _move(self):
    self.offset = self.offset + self.speed
    self._calc_viewport_index()
    if not self.speed: return
    self._apply_friction()

  def _update_sprite_rect(self,index,offset):
    sprite = self.sorted_sprites[index]
    sprite.rect.top  = self.rect.top + offset
    sprite.rect.left = self.rect.left
    if not self.rect.colliderect(sprite.rect): return False
    return True

  def is_index_mostly_visible(self,index):
    if not self.sorted_sprites: return False
    i = self.viewport_index
    if self.viewport_index == index: first = False
    else: first = True
    c = 0
    while self.rect.collidepoint(self.sorted_sprites[i].rect.center) or first:
      if i == index: return True
      c += 1
      i = self._calc_new_index(i+1,self.sorted_sprites)
      if i == self.viewport_index: break # handle lists that are shorter than the viewport
      first = False
    return False

  def update(self):
    self._move()
    self._sort_sprites()
    if not self.sorted_sprites: return

    i      = self.viewport_index
    offset = self.offset
    first  = True
    while self._update_sprite_rect(i,offset) or first:
      offset = self.sorted_sprites[i].rect.bottom + self.y_spacer - self.rect.top
      i = self._calc_new_index(i+1,self.sorted_sprites)
      first = False
      if i == self.viewport_index: break # handle lists that are shorter than the viewport

    if self.has_selected():
      self.get_selected_sprite().update()
      self.selection_indicator.rect.centery = self.get_selected_sprite().rect.centery
      self.selection_indicator.rect.left = self.rect.left - self.selection_indicator.rect.width - 2
    if self.is_scrollable == None: self.is_scrollable = self._is_scrollable()

  def _is_tap(self):
    if self.num_events <= \
        self.tap_event_threshold and abs(self.distance) <= \
        self.tap_distance_threshold: return True
    return False

  def tap_select(self,run_callback=True):
    if not self.sorted_sprites: return
    mouse_pos = pygame.mouse.get_pos()
    i      = self.viewport_index
    first  = True
    while self.rect.colliderect(self.sorted_sprites[i].rect) or first:
      if self.sorted_sprites[i].rect.collidepoint(mouse_pos):
        self.set_selected_index(i)
        if run_callback:
          self.selected_cb()
          self._change_state('idle')
        break
      i = self._calc_new_index(i+1,self.sorted_sprites)
      first = False
      if i == self.viewport_index: break # handle lists that are shorter than the viewport

  def _is_scrollable(self):
    if not self.sorted_sprites: return False
    i      = self.viewport_index
    first  = True
    while self.rect.colliderect(self.sorted_sprites[i].rect) or first:
      i = self._calc_new_index(i+1,self.sorted_sprites)
      first = False
      if i == self.viewport_index: return False # handle lists that are shorter than the viewport
    return True

  def handle_event(self,event):
    if event.type == MOUSEBUTTONDOWN:
      if self.rect.collidepoint(pygame.mouse.get_pos()):
        manager.has_focus = True
        self.tap_select(False)
        self.speed      = 0
        self.distance   = 0
        self.num_events = 0
        self.mousedown = 1
    elif event.type == MOUSEBUTTONUP:
      if self.mousedown:
        self.mousedown  = 0
        self.friction_c = 0
        if self._is_tap():
          # This was a tap, not a drag
          self._change_state('tap')
          self.stop_animation()
    elif event.type == MOUSEMOTION:
      if self.mousedown and self.sorted_sprites and self.is_scrollable:
        (rel_x,rel_y) = event.rel
        self.offset = self.offset + rel_y
        if cmp(rel_y,0) and cmp(rel_y,0) != cmp(self.distance,0): # changed direction
          self.num_events = 0
          self.distance   = 0
        self.num_events = self.num_events + 1
        self.distance   = self.distance   + rel_y
        self.speed      = self.sorted_sprites[0].rect.height / 50 * self.distance / self.num_events
        if not self._is_tap(): self._change_state('drag')
    elif event.type == KEYDOWN and event.key == K_RETURN:
      if not self.has_selected(): self.new_selected()
      self.selected_cb()

    if self.last_state == 'idle' and self.cur_state == 'tap': self.tap_select()

  def is_idle(self):
    isidle = (not self.speed and not self.mousedown)
    if isidle: self._change_state('idle')
    return isidle

  def draw(self,surface):
    if not self.sorted_sprites: return
    i = self.viewport_index
    first = True
    while first or self.rect.colliderect(self.sorted_sprites[i].rect):
      self.sorted_sprites[i].draw(surface)
      i = self._calc_new_index(i+1,self.sorted_sprites)
      first = False
      if i == self.viewport_index: break # handle lists that are shorter than the viewport
    if self.has_selected() and self.is_index_mostly_visible(self.selected_index): self.selection_indicator.draw(surface)

  def stop_animation(self):
    self.speed      = 0
    self.distance   = 0
    self.num_events = 0
    self.mousedown  = 0
    self.friction_c = 0
    self.update()

  def close(self):
    for widget in self.sorted_sprites: widget.remove()


class Album(BaseWidget):
  def __init__(self,name,show_reflection=True):
    BaseWidget.__init__(self)
    self.name       = name
    self.show_reflection = show_reflection
    self.theme_rect = theme.get_rect('album_art_bg')
    if not self.show_reflection: self.theme_rect.height = 125
    self.cache_rect = None
    self.rect       = None
    self.get_cache_rect()
    self.rect       = Rect(0,0,self.theme_rect.width,self.theme_rect.height)

  def get_cache_rect(self):
    list = db.get_album_art(self.name)
    if not list:
      print "invalid album cache for %s" % (self.name,)
      return # should never happen

    row = list[0]
    self.cache_rect = Rect(row['x'],row['y'],self.theme_rect.width,self.theme_rect.height)

  def draw(self,surface,point=None):
    if not point: point = self.rect
    surface.blit(globals.albumcache,point,self.cache_rect)

  def __cmp__(self,other):
    return cmp(self.name.lower(),other.name.lower())


class PlayButton(PlayBackButton):
  def __init__(self):
    ''' I really hope I don't have to explain this button. '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'play')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 675 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 160
    self.playing   = False

  def _load_image(self):
    if manager.playlist.is_paused(): theme.blit(self.image,(50,0),'play')
    else: theme.blit(self.image,(50,0),'pause')

  def update(self):
    self._load_image()

  def _make_callback(self):
    def _callback():
      manager.playlist.pause()
      self.update()
      return 
    return _callback


class NextButton(PlayBackButton):
  def __init__(self):
    ''' Skips forward by one track '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'next')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 675 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 225

  def _make_callback(self):
    def _callback():
      manager.playlist.next(None, None)
      return 
    return _callback


class PrevButton(PlayBackButton):
  def __init__(self):
    ''' Skips backward by one track '''
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(50,0),'prev')
    PlayBackButton.__init__(self,self.image,self._make_callback())
    self.rect.left = 675 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = 100

  def _make_callback(self):
    def _callback():
      manager.playlist.prev(None, None)
      return 
    return _callback


class RepeatButton(Button):
  def __init__(self):
    ''' Repeat: (none|one|all) '''

    self.image_x = 27
    self.image = theme.get_surface((125,43))
    theme.blit(self.image,(self.image_x,0),'repeat_off')
    Button.__init__(self,self.image,self._make_callback())

    self.rect.left = 0 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = manager.screen.get_rect().height - self.image.get_rect().height - 110
    self.read_interpret_mode()
    self.update()

  def _make_callback(self):
    def _callback():
      if self.mode=='repeat_off': self.mode='repeat_all'
      elif self.mode=='repeat_all': self.mode='repeat_one'
      elif self.mode=='repeat_one': self.mode='repeat_shuffle'
      else: self.mode='repeat_off'
      self.update()
      self.write_mode()
      print "RepeatButton: new mode is %s" % self.mode
      return 
    return _callback

  def read_interpret_mode(self):
    if manager.playlist.random: self.mode='repeat_shuffle'
    elif manager.playlist.repeat_one: self.mode='repeat_one'
    elif manager.playlist.repeat: self.mode='repeat_all'
    else: self.mode='repeat_off'

  def write_mode(self):
    if self.mode=='repeat_one':
      manager.playlist.repeat=True
      manager.playlist.repeat_one=True
      manager.playlist.continuous=True
      manager.playlist.random=False
    elif self.mode=='repeat_all':
      manager.playlist.repeat=True
      manager.playlist.repeat_one=False
      manager.playlist.continuous=True
      manager.playlist.random=False
    elif self.mode=='repeat_shuffle':
      manager.playlist.repeat=True
      manager.playlist.repeat_one=False
      manager.playlist.continuous=True
      manager.playlist.random=True
    else:
      manager.playlist.repeat=False
      manager.playlist.repeat_one=False
      manager.playlist.continuous=True
      manager.playlist.random=False

  def update(self):
    theme.blit(self.image,(self.image_x,0),self.mode) # FIXME: this is copy/paste without thinking, think/fix


class A2DPButton(Button):
  def __init__(self):
    ''' This is the bluetooth button in the upper right. It displays
    the status of bluez's a2dpd. '''
    self.image = theme.get_surface((125,43))
    Button.__init__(self,self.image,self._make_a2dp_button_callback())
    if globals.no_bt_mplayer or (not manager.has_a2dp and not globals.theme_tester): return
    self.tester_a2dp_status = False
    theme.blit(self.image,(45,0),'a2dp_off')
    self.rect.left = 675
    self.rect.top  = 10

  def _load_image(self):
    if globals.theme_tester:
      a2dp = self.tester_a2dp_status
    else:
      import mplayer
      if isinstance(manager.player, mplayer.Mplayer):
        a2dp = manager.player.get_a2dpd()
      else: a2dp = False
    print "a2dp: %s" % (a2dp,)
    self.image.fill(theme.colorkey)
    if a2dp: theme.blit(self.image,(45,0),'a2dp_on')
    else: theme.blit(self.image,(45,0),'a2dp_off')

  def update(self):
    if globals.no_bt_mplayer or (not manager.has_a2dp and not globals.theme_tester): return
    self._load_image()

  def _make_a2dp_button_callback(self):
    def _a2dp_button_callback():
      print "switching a2dp modes from button"
      manager.clear()
      status_area = Text("PLEASE WAIT...")
      status_area.rect.topleft = (130,200)
      status_area.update()
      status_area.draw(manager.screen)
      pygame.display.flip()
      
      if globals.theme_tester:
        self.tester_a2dp_status = not self.tester_a2dp_status
        self.update()
        return
      
      import mplayer
      if not isinstance(manager.player, mplayer.Mplayer):
        # switch to mplayer for a2dp
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        seconds = manager.player.seconds
        manager.player.close()
        manager.player = mplayer.Mplayer()
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
          manager.player.seek(seconds,2,True)
        if target and paused: manager.player.pause(True)
        if target and seconds: # make sure we're currently listening to something
          while not manager.player.seconds:
            manager.player.queryStatus()
            time.sleep(0.1)
      else:
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        seconds = manager.player.seconds
      manager.player.set_a2dpd(not manager.player.get_a2dpd())
      if not manager.player.get_a2dpd():
        # switch to normal player for non a2dp
        manager.player.close()
        manager._get_player()
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
          manager.player.seek(seconds,2,True)
        if target and paused: manager.player.pause(True)
      self.update()
      return 
    return _a2dp_button_callback


class PlayerButton(Button):
  def __init__(self):
    ''' This is the bluetooth button in the upper right. It displays
    the current active player: OSSOPlayer or MPlayer '''
    self.image = theme.get_surface((125,43))
    Button.__init__(self,self.image,self._make_player_button_callback())
    if not globals.no_bt_mplayer or not manager.has_mplayer: return
    theme.blit(self.image,(45,0),'a2dp_off')
    self.rect.left = 675
    self.rect.top  = 10

  def _load_image(self):
    import mplayer
    if isinstance(manager.player, mplayer.Mplayer):
      mpl = True
    else: mpl = False
    print "use mplayer: %s" % (mpl,)
    self.image.fill(theme.colorkey)
    if mpl: theme.blit(self.image,(45,0),'a2dp_on')
    else: theme.blit(self.image,(45,0),'a2dp_off')

  def update(self):
    if not globals.no_bt_mplayer or not manager.has_mplayer: return
    self._load_image()

  def _make_player_button_callback(self):
    def _player_button_callback():
      print "switching players from button"
      manager.clear()
      status_area = Text("PLEASE WAIT...")
      status_area.rect.topleft = (130,200)
      status_area.update()
      status_area.draw(manager.screen)
      pygame.display.flip()
      
      import mplayer
      if not isinstance(manager.player, mplayer.Mplayer):
        # switch to mplayer
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        seconds = manager.player.seconds
        manager.player.close()
        manager.player = mplayer.Mplayer()
        manager.player.set_a2dpd(False)
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
          manager.player.seek(seconds,2,True)
        if target and paused: manager.player.pause(True)
        if target and seconds: # make sure we're currently listening to something
          while not manager.player.seconds:
            manager.player.queryStatus()
            time.sleep(0.1)
      else:
        target = manager.player.target
        paused = manager.player.paused
        volume = manager.player.volume
        seconds = manager.player.seconds
        manager.player.close()
        import ossoplayer
        manager.player = ossoplayer.OSSOPlayer()
        if target:
          manager.player.play(target)
          manager.player.set_volume(volume,show_volumeview=False)
          manager.player.seek(seconds,2,True)
        if target and paused: manager.player.pause(True)
      self.update()
      return 
    return _player_button_callback


class NowPlayingButton(Button):
  def __init__(self):
    ''' This is the button in the lower right. It displays
    the currently playing album and song. '''
    self.title_text_x = 2
    self.title_text_y = 152
    self.position_text_x = 2
    self.position_text_y = 0
    self.album_x = 0
    self.album_y = 20
    self.image = theme.get_surface((125,160+self.album_y))
    self.pbar_background = None
    theme.blit(self.image,(self.album_x,self.album_y),'album_art_bg')
    Button.__init__(self,self.image,self._make_nowplaying_button_callback())
    self.rect.left = 675 # FIXME: the controlling view should probably set the x,y pos for us
    self.rect.top  = manager.screen.get_rect().height - self.image.get_rect().height
    self.current   = None
    self.last_album_sprite = None
    self.position_string = ""
    self.current_song = None
    self.pbar         = None
    self._get_pbar_background()

  def _get_pbar_background(self):
    self.pbar_background = theme.get_surface((self.rect.width,self.album_y - self.position_text_y))
    self.pbar_background.blit(manager.background,(0,0),Rect(
      self.rect.left,
      self.rect.top  + self.position_text_y,
      self.rect.width,
      self.album_y - self.position_text_y))
    self.pbar_background.fill((0,0,0),Rect(0,self.pbar_background.get_rect().height - 5,self.pbar_background.get_rect().width,5))

  def _load_image(self):
    # position text
    if self.current_song and manager.playlist.status() == self.current:
      song = self.current_song
    else:
      song = self.current_song = db.song_of_path(manager.playlist.status())[0]
    if not self.pbar: self.pbar = ProgressBar(
        Rect(0,self.album_y - 4,self.rect.width,3),segmented=False)

    current_time  = globals.format_time(manager.player.seconds)
    player_length = globals.format_time(manager.player.length)
    new_pos_s     = "%s of %s (%3d%%)" % (current_time,player_length,manager.player.percent)
    if self.position_string != new_pos_s:
      self.pbar.set_percent(manager.player.percent)
      self.position_string = new_pos_s
      self.image.blit(self.pbar_background,(0,self.position_text_y))
      self.pbar.blit(self.image)
      globals.gprint(
        self.position_string,
        self.image,
        (self.position_text_x, self.position_text_y),
        16,
        False)
    if manager.playlist.status() == self.current: return
    print "nowplaying: " + song['album']
    if not self.last_album_sprite or self.last_album_sprite.name != song['album']:
      self.last_album_sprite = Album(song['album'])
    self.last_album_sprite.draw(self.image,(self.album_x,self.album_y))

    track = song['track']
    if not track: track = ''
    else: track = "%02d " % (track,)
    # title text
    globals.gprint(
      track+song['title'],
      self.image,
      (self.title_text_x, self.title_text_y),
      16)
    self.current = manager.playlist.status()

  def update(self):
    if not manager.playlist.status():
      return
    self._load_image()

  def _make_nowplaying_button_callback(self):
    def _nowplaying_button_callback():
      print "show now_playing"
      manager.view.scroll_widget.stop_animation()
      return manager.show_nowplaying()
    return _nowplaying_button_callback


class AlbumArtist(BaseWidget):
  def __init__(self,album,artist,selected=False):
    BaseWidget.__init__(self)
    self.selected = selected
    self.theme_rect = None
    self.get_theme_rect()
    self.rect     = Rect(0,0,self.theme_rect.width,self.theme_rect.height)
    self.album    = Album(album,show_reflection=False)
    self.artist   = RawText(self.get_artist_rect(),artist,album)

  def get_artist_rect(self):
    x = self.album.rect.right
    y = self.rect.top + 30
    return Rect(x,y,self.rect.right - 15 - x, self.rect.bottom - y)

  def get_theme_rect(self):
    if self.selected:
      self.theme_rect = theme.get_rect('albumartist_bg_sel')
    else:
      self.theme_rect = theme.get_rect('albumartist_bg')

  def draw(self,surface):
    surface.blit(theme.surface,self.rect,self.theme_rect)
    self.album.rect.topleft = (self.rect.left + 16,self.rect.top + 6)
    self.album.draw(surface)
    self.artist.rect.topleft = self.get_artist_rect().topleft
    self.artist.draw(surface)

  def update(self):
    self.album.update()
    self.artist.update()

  def set_focus(self,focus):
    self.artist.set_focus(focus)
      
  def __cmp__(self,other):
    return cmp(
        ': '.join((self.artist.name.lower(),self.album.name.lower())),
        ': '.join((other.artist.name.lower(),other.album.name.lower()))
    )


# This is **not** a widget. It's more of a font wrapper. It makes it easy to implement scrolling text within a rectangle.
class ScrollingText():
  def __init__(self,name,rect,size=False):
    self.rect     = rect
    self.name     = name
    self.size     = size
    self.offset   = 0
    self.step     = 10  # number of pixels to move text
    self.frequency= 0.1 # how often to move text, in seconds
    self.focus    = False
    self.font_surf= globals.gprint_ret(self.name,self.size)
    self.sep_font_surf= globals.gprint_ret(' ~ ',self.size)
    self.scroll_timeout_start = 0

  def draw(self,surface):
    old_clip = surface.get_clip()
    surface.set_clip(self.rect)
    surface.blit(self.font_surf,(self.rect.left-self.offset,self.rect.top))
    if self.offset:
      surface.blit(self.sep_font_surf,(self.rect.left-self.offset + self.font_surf.get_rect().width,self.rect.top))
      surface.blit(self.font_surf,(self.rect.left-self.offset + self.font_surf.get_rect().width + self.sep_font_surf.get_rect().width,self.rect.top))
    surface.set_clip(old_clip)
      
  def set_focus(self,focus):
    print "set_focus: %s for %s" % (focus,self.name)
    if focus != self.focus:
      if not focus:
        self.offset = 0
    self.focus = focus

  def update(self):
    if not self.scroll_timeout_start or time.time() - self.scroll_timeout_start >= self.frequency:
      self.scroll_timeout_start = time.time()
      if self.font_surf.get_rect().width> self.rect.width:
        self.offset = self.offset + self.step
        if self.offset >= self.font_surf.get_rect().width + self.sep_font_surf.get_rect().width: self.offset = 0
        return True
    return False


class RawText(BaseWidget):
  def __init__(self, rect, name, subtext=None, selected=False):
    BaseWidget.__init__(self)
    self.name     = name
    self.subtext  = subtext
    self.selected = selected
    self.rect     = rect
    self.focus    = False
    self.scrolling_text = None
    self.scrolling_subtext = None
    self.text_x   = 15
    self.text_y   = 25
    if self.subtext: self.text_y = self.text_y - 15
    self.get_scrolling_text()
    self.get_scrolling_subtext()

  def get_scrolling_text(self):
    if not self.scrolling_text:
      self.scrolling_text = ScrollingText(
          self.name,
          Rect(self.text_x, self.text_y, self.rect.width - (self.text_x * 2) - self.text_x, self.rect.height))

  def get_scrolling_subtext(self):
    if self.subtext and not self.scrolling_subtext:
      self.scrolling_subtext = ScrollingText(
          self.subtext,
          Rect(self.text_x, self.text_y+30, self.rect.width - (self.text_x * 2) - self.text_x, self.rect.height),
          30)

  def draw(self,surface):
    self.scrolling_text.rect.topleft = (self.rect.left + self.text_x, self.rect.top + self.text_y)
    self.scrolling_text.draw(surface)
    if self.scrolling_subtext:
      self.scrolling_subtext.rect.topleft = (self.rect.left + self.text_x, self.rect.top + self.text_y+30)
      self.scrolling_subtext.draw(surface)

  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.name.lower(),other.name.lower())

  def set_selected(self,selected):
    if selected != self.selected: # ignore no change
      self.selected = selected

  def set_focus(self,focus):
    self.scrolling_text.set_focus(focus)
    if self.scrolling_subtext: self.scrolling_subtext.set_focus(focus)
    print "set_focus: %s for %s" % (focus,self.name)

  def update(self):
    self.scrolling_text.update()
    if self.scrolling_subtext: self.scrolling_subtext.update()


class Text(RawText):
  def __init__(self, name, subtext=None, selected=False):
    self.selected = selected
    self.theme_rect = None
    self.get_theme_rect()
    self.rect     = Rect(0,0,self.theme_rect.width,self.theme_rect.height)
    RawText.__init__(self,self.rect,name,subtext,self.selected)

  def get_theme_rect(self):
    if self.selected:
      self.theme_rect = theme.get_rect('text_bg_sel')
    else:
      self.theme_rect = theme.get_rect('text_bg')

  def draw(self,surface):
    surface.blit(theme.surface,self.rect,self.theme_rect)
    RawText.draw(self,surface)

  def set_selected(self,selected):
    if selected != self.selected: # ignore no change
      self.selected = selected
      self.get_theme_rect()


class Song(Text):
  def __init__(self,track,title,path):
    self.track    = self._fix_track(track)
    self.title    = title
    self.path     = path
    self.track_font_surf= globals.gprint_ret(self.track)
    Text.__init__(self,self.title)

  def _fix_track(self,track):
    if not track: track = ''
    else: track = "%02d " % (track,)
    return track

  def get_scrolling_text(self):
    if not self.scrolling_text:
      self.scrolling_text = ScrollingText(
          self.name,
          Rect(
            self.text_x + self.track_font_surf.get_rect().width,
            self.text_y,
            self.rect.width - (self.text_x * 2) - self.track_font_surf.get_rect().width,
            self.rect.height))

  def draw(self,surface):
    surface.blit(theme.surface,self.rect,self.theme_rect)
    surface.blit(self.track_font_surf,(self.rect.left + self.text_x,self.rect.top + self.text_y))
    self.scrolling_text.rect.topleft = (self.rect.left + self.text_x + self.track_font_surf.get_rect().width, self.rect.top + self.text_y)
    self.scrolling_text.draw(surface)
    if self.scrolling_subtext: self.scrolling_subtext.draw(surface)

  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.track + self.name.lower(),other.track + other.name.lower())


# This is **not** a widget. It's more of a blit helper. It makes it easy to implement progress bar within a rectangle.
class ProgressBar():
  def __init__(self, rect, percent=0, color=None, segmented=True):
    ''' Displays a graphical bar representing a positive integer 
    between 0 and 100, inclusive '''
    self.rect      = rect
    self.percent   = int(percent)
    self.segmented = segmented
    if color == None: self.bar_color = theme.bar_color_sel
    else: self.bar_color = color

  def blit(self,image):
    max_bar_width = self.rect.right - self.rect.left
    current_bar_width = max_bar_width * self.percent / 100
    bar_segment_width = max_bar_width / 10
    if not self.segmented: image.fill(self.bar_color,Rect(self.rect.left,self.rect.top,current_bar_width,self.rect.bottom - self.rect.top))
    else:
      c = 0
      while c < current_bar_width:
        if c + bar_segment_width < current_bar_width:
          current_segment_width = bar_segment_width
        else:
          current_segment_width = current_bar_width - c
        image.fill(self.bar_color,Rect(self.rect.left + c,self.rect.top,current_segment_width - 1,self.rect.bottom - self.rect.top))
        c += bar_segment_width
      
  def set_color(self,color):
    self.bar_color = color

  def set_percent(self,percent):
    self.percent = int(percent)


class PercentBar(BaseWidget):
  def __init__(self,percent, selected=False):
    ''' Displays a positive integer between 0 and 100, inclusive
    along with a graphical bar representing this number '''
    BaseWidget.__init__(self)
    self.pbar      = None
    self.rect      = None
    self.percent   = 0
    self.font_surf = None
    self.selected  = selected
    self.set_percent(percent)

  def _load_image(self):
    if self.selected:
      self.image = theme.get_image('text_bg_sel')
      bar_color  = theme.bar_color_sel
    else:
      self.image = theme.get_image('text_bg')
      bar_color  = theme.bar_color

    offset = self.font_surf.get_rect().width + 5 + 5
    self.image.blit(self.font_surf,(5,24))
    if not self.rect: self.rect = self.image.get_rect()
    if not self.pbar: self.pbar = ProgressBar(
        Rect(offset,24,self.rect.right - self.rect.left - offset - 15,30),
        self.percent,
        bar_color)
    else:
      self.pbar.set_color(bar_color)
      self.pbar.set_percent(self.percent)
    self.pbar.blit(self.image)
      
  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.percent,other.percent)

  def set_selected(self,selected):
    if selected != self.selected: # ignore no change
      self.selected = selected
      self._load_image()

  def set_focus(self,focus):
    ''' do nothing '''

  def set_percent(self,percent):
    self.percent = int(percent)
    self.font_surf = globals.gprint_ret("%3d%%" % (self.percent,))
    offset = self.font_surf.get_rect().width + 5 + 5
    self.pbar = None
    self._load_image()


class Time(Text):
  def __init__(self,seconds):
    self.seconds  = seconds
    Text.__init__(self,globals.format_time(self.seconds))

  def __cmp__(self,other):
    if not other: return 0
    return cmp(self.seconds,other.seconds)


class Bar(BaseWidget):
  def __init__(self,rect):
    BaseWidget.__init__(self)
    self.rect  = Rect(rect)
    self.image = theme.get_surface((self.rect.width,self.rect.height))
